// Copyright 2021 Matrix Origin
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package meta

import (
	dbsched "matrixone/pkg/vm/engine/aoe/storage/db/sched"
	"matrixone/pkg/vm/engine/aoe/storage/dbi"
	md "matrixone/pkg/vm/engine/aoe/storage/metadata/v1"
	"matrixone/pkg/vm/engine/aoe/storage/testutils/config"
	"os"
	"path"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
)

var (
	workDir = "/tmp/mevents"
)

func TestBasicOps(t *testing.T) {
	opts := config.NewOptions(workDir, config.CST_Customize, config.BST_S, config.SST_S)
	opts.Scheduler = dbsched.NewScheduler(opts, nil)

	now := time.Now()

	info := opts.Meta.Info
	tabletInfo := md.MockTableInfo(2)
	eCtx := &dbsched.Context{Opts: opts, Waitable: true}
	createTblE := NewCreateTableEvent(eCtx, dbi.TableOpCtx{TableName: tabletInfo.Name}, tabletInfo)
	opts.Scheduler.Schedule(createTblE)
	err := createTblE.WaitDone()
	assert.Nil(t, err)

	tbl := createTblE.GetTable()
	assert.NotNil(t, tbl)

	t.Log(info.String())
	createBlkE := NewCreateBlkEvent(eCtx, tbl.ID, nil)
	opts.Scheduler.Schedule(createBlkE)
	err = createBlkE.WaitDone()
	assert.Nil(t, err)

	blk1 := createBlkE.GetBlock()
	assert.NotNil(t, blk1)
	assert.Equal(t, blk1.GetBoundState(), md.Detached)

	assert.Equal(t, blk1.DataState, md.EMPTY)
	blk1.SetCount(blk1.MaxRowCount)
	assert.Equal(t, blk1.DataState, md.FULL)

	blk2, err := info.ReferenceBlock(blk1.Segment.Table.ID, blk1.Segment.ID, blk1.ID)
	assert.Nil(t, err)
	assert.Equal(t, blk2.DataState, md.EMPTY)
	assert.Equal(t, blk2.Count, uint64(0))

	schedCtx := &dbsched.Context{
		Opts:     opts,
		Waitable: true,
	}
	commitCtx := &dbsched.Context{Opts: opts, Waitable: true}
	commitCtx.AddMetaScope()
	commitE := dbsched.NewCommitBlkEvent(commitCtx, blk1)
	opts.Scheduler.Schedule(commitE)
	err = commitE.WaitDone()
	assert.Nil(t, err)

	blk3, err := info.ReferenceBlock(blk1.Segment.Table.ID, blk1.Segment.ID, blk1.ID)
	assert.Nil(t, err)
	assert.Equal(t, blk3.DataState, md.FULL)
	assert.Equal(t, blk1.Count, blk3.Count)

	for i := 0; i < 100; i++ {
		createBlkE = NewCreateBlkEvent(schedCtx, blk1.Segment.Table.ID, nil)
		opts.Scheduler.Schedule(createBlkE)
		err = createBlkE.WaitDone()
		assert.Nil(t, err)
	}
	du := time.Since(now)
	t.Log(du)

	ctx := md.CopyCtx{Attached: true}
	info_copy := info.Copy(ctx)

	fpath := path.Join(info.Conf.Dir, "tttttttt")
	w, err := os.Create(fpath)
	assert.Equal(t, err, nil)
	err = info_copy.Serialize(w)
	assert.Nil(t, err)
	w.Close()

	// r, err := os.OpenFile(fpath, os.O_RDONLY, 0666)
	// assert.Equal(t, err, nil)

	// de_info, err := md.Deserialize(r)
	// assert.Nil(t, err)
	// assert.NotNil(t, de_info)
	// assert.Equal(t, info.Sequence.NextBlockID, de_info.Sequence.NextBlockID)
	// assert.Equal(t, info.Sequence.NextSegmentID, de_info.Sequence.NextSegmentID)
	// assert.Equal(t, info.Sequence.NextTableID, de_info.Sequence.NextTableID)

	// r.Close()

	du = time.Since(now)
	t.Log(du)

	opts.Scheduler.Stop()
}
